package ru.dbotthepony.kstarbound.client.world

import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.longs.LongArraySet
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ReferenceArraySet
import ru.dbotthepony.kstarbound.PIXELS_IN_STARBOUND_UNITf
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.render.ConfiguredMesh
import ru.dbotthepony.kstarbound.client.render.LayeredRenderer
import ru.dbotthepony.kstarbound.client.render.Mesh
import ru.dbotthepony.kstarbound.client.render.RenderLayer
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.math.roundTowardsNegativeInfinity
import ru.dbotthepony.kstarbound.math.roundTowardsPositiveInfinity
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.NeverFilter
import ru.dbotthepony.kstarbound.world.NonEmptyFilter
import ru.dbotthepony.kstarbound.world.RayCastResult
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.api.ITileAccess
import ru.dbotthepony.kstarbound.world.api.OffsetCellAccess
import ru.dbotthepony.kstarbound.world.api.TileView
import ru.dbotthepony.kstarbound.world.castRay
import ru.dbotthepony.kstarbound.world.positiveModulo
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector2f
import ru.dbotthepony.kvector.vector.Vector2i
import kotlin.math.PI
import kotlin.math.cos
import kotlin.math.sin

class ClientWorld(
	val client: StarboundClient,
	seed: Long,
	size: Vector2i? = null,
	loopX: Boolean = false,
	loopY: Boolean = false
) : World<ClientWorld, ClientChunk>(seed, size, loopX, loopY) {
	init {
		physics.debugDraw = client.box2dRenderer
	}

	private fun determineChunkSize(cells: Int): Int {
		for (i in 32 downTo 1) {
			if (cells % i == 0) {
				return i
			}
		}

		throw RuntimeException("unreachable code")
	}

	val renderRegionWidth = if (size == null) 16 else determineChunkSize(size.x)
	val renderRegionHeight = if (size == null) 16 else determineChunkSize(size.y)
	val renderRegionsX = if (size == null) 0 else size.x / renderRegionWidth
	val renderRegionsY = if (size == null) 0 else size.y / renderRegionHeight

	fun isValidRenderRegionX(value: Int): Boolean {
		if (size == null || loopX) {
			return true
		} else {
			return value in 0 .. renderRegionsX
		}
	}

	fun isValidRenderRegionY(value: Int): Boolean {
		if (size == null || loopY) {
			return true
		} else {
			return value in 0 .. renderRegionsY
		}
	}

	inner class RenderRegion(val x: Int, val y: Int) {
		inner class Layer(private val view: ITileAccess, private val isBackground: Boolean) {
			val bakedMeshes = ArrayList<Pair<ConfiguredMesh<*>, RenderLayer.Point>>()
			var isDirty = true

			fun bake() {
				if (!isDirty) return
				isDirty = false

				bakedMeshes.clear()

				val meshes = LayeredRenderer()

				for (x in 0 until renderRegionWidth) {
					for (y in 0 until renderRegionHeight) {
						if (!inBounds(x, y)) continue

						val tile = view.getTile(x, y) ?: continue
						val material = tile.material

						if (!material.isMeta) {
							client.tileRenderers
								.getMaterialRenderer(material.materialName)
								.tesselate(tile, view, meshes, Vector2i(x, y), isBackground = isBackground)
						}

						val modifier = tile.modifier

						if (modifier != null) {
							client.tileRenderers
								.getModifierRenderer(modifier.modName)
								.tesselate(tile, view, meshes, Vector2i(x, y), isBackground = isBackground, isModifier = true)
						}
					}
				}

				for ((baked, zLevel) in meshes.bakeIntoMeshes()) {
					bakedMeshes.add(baked to zLevel)
				}
			}
		}

		private val liquidMesh = ArrayList<Pair<Mesh, RGBAColor>>()
		var liquidIsDirty = true

		val view = OffsetCellAccess(this@ClientWorld, x * renderRegionWidth, y * renderRegionHeight)

		val background = Layer(TileView.Background(view), true)
		val foreground = Layer(TileView.Foreground(view), false)

		fun addLayers(layers: LayeredRenderer, renderOrigin: Vector2f) {
			background.bake()
			foreground.bake()

			if (liquidIsDirty) {
				liquidIsDirty = false
				liquidMesh.clear()

				val liquidTypes = ReferenceArraySet<LiquidDefinition>()

				for (x in 0 until renderRegionWidth) {
					for (y in 0 until renderRegionHeight) {
						view.getCell(x, y)?.liquid?.def?.let { liquidTypes.add(it) }
					}
				}

				for (type in liquidTypes) {
					val builder = client.programs.liquid.builder.builder

					builder.begin()

					for (x in 0 until renderRegionWidth) {
						for (y in 0 until renderRegionHeight) {
							val state = view.getCell(x, y)

							if (state?.liquid?.def === type) {
								builder.vertex(x.toFloat(), y.toFloat())
								builder.vertex(x.toFloat() + 1f, y.toFloat())
								builder.vertex(x.toFloat() + 1f, y.toFloat() + 1f)
								builder.vertex(x.toFloat(), y.toFloat() + 1f)
							}
						}
					}

					liquidMesh.add(Mesh(builder) to type.color)
				}
			}

			for ((baked, zLevel) in background.bakedMeshes) {
				layers.add(zLevel) {
					client.stack.push().last().translate(renderOrigin.x, renderOrigin.y)
					baked.render()
					client.stack.pop()
				}
			}

			for ((baked, zLevel) in foreground.bakedMeshes) {
				layers.add(zLevel) {
					client.stack.push().last().translate(renderOrigin.x, renderOrigin.y)
					baked.render()
					client.stack.pop()
				}
			}

			/*if (liquidMesh.isNotEmpty()) {
				layers.add(RenderLayer.Liquid.index) {
					client.stack.push().last().translate(renderOrigin.x, renderOrigin.y)

					val program = client.programs.liquid

					program.use()
					program.transform = client.stack.last()

					for ((mesh, color) in liquidMesh) {
						program.baselineColor = color
						mesh.render()
					}

					client.stack.pop()
				}
			}*/
		}
	}

	val renderRegions = Long2ObjectOpenHashMap<RenderRegion>()

	fun renderRegionKey(x: Int, y: Int): Long {
		if (size == null) {
			return x.toLong() shl 32 or y.toLong()
		} else {
			return positiveModulo(x, renderRegionsX).toLong() shl 32 or positiveModulo(y, renderRegionsY).toLong()
		}
	}

	/**
	 * all intersecting chunks
	 */
	inline fun forEachRenderRegion(pos: ChunkPos, action: (RenderRegion) -> Unit) {
		var (ix, iy) = pos.tile
		ix /= renderRegionWidth
		iy /= renderRegionHeight

		for (x in ix .. ix + CHUNK_SIZE / renderRegionWidth) {
			for (y in iy .. iy + CHUNK_SIZE / renderRegionWidth) {
				renderRegions[renderRegionKey(x, y)]?.let(action)
			}
		}
	}

	/**
	 * cell and cells around
	 */
	inline fun forEachRenderRegion(pos: IStruct2i, action: (RenderRegion) -> Unit) {
		val dx = pos.component1() % renderRegionWidth
		val dy = pos.component2() % renderRegionHeight

		if (dx == 1 || dx == renderRegionWidth - 1 || dy == 1 || dy == renderRegionHeight - 1) {
			val seen = LongArraySet(8)

			for ((x, y) in ring) {
				val ix = (pos.component1() + x) / renderRegionWidth
				val iy = (pos.component2() + y) / renderRegionHeight
				val index = renderRegionKey(ix, iy)

				if (seen.add(index)) {
					renderRegions[index]?.let(action)
				}
			}
		} else {
			val ix = pos.component1() / renderRegionWidth
			val iy = pos.component2() / renderRegionHeight
			renderRegions[renderRegionKey(ix, iy)]?.let(action)
		}
	}

	override fun chunkFactory(pos: ChunkPos): ClientChunk {
		return ClientChunk(this, pos)
	}

	fun addLayers(
		size: AABB,
		layers: LayeredRenderer
	) {
		val rx = roundTowardsNegativeInfinity(size.mins.x) / renderRegionWidth - 1
		val ry = roundTowardsNegativeInfinity(size.mins.y) / renderRegionHeight - 1

		val dx = roundTowardsPositiveInfinity(size.maxs.x - size.mins.x) / renderRegionWidth + 2
		val dy = roundTowardsPositiveInfinity(size.maxs.y - size.mins.y) / renderRegionHeight + 2

		for (x in rx .. rx + dx) {
			for (y in ry .. ry + dy) {
				if (!isValidRenderRegionX(x) || !isValidRenderRegionY(y)) continue
				val renderer = renderRegions.computeIfAbsent(renderRegionKey(x, y), Long2ObjectFunction {
					RenderRegion((it ushr 32).toInt(), it.toInt())
				})

				renderer.addLayers(layers, Vector2f(x * renderRegionWidth.toFloat(), y * renderRegionHeight.toFloat()))
			}
		}

		for (obj in objects) {
			if (obj.pos.x in client.viewportCellX .. client.viewportCellX + client.viewportCellWidth && obj.pos.y in client.viewportCellY .. client.viewportCellY + client.viewportCellHeight) {
				val layer = layers.getLayer(obj.orientation?.renderLayer ?: continue)

				obj.drawables.forEach {
					val (x, y) = obj.imagePosition
					it.render(client, layer, obj.pos.x.toFloat() + x / PIXELS_IN_STARBOUND_UNITf, obj.pos.y.toFloat() + y / PIXELS_IN_STARBOUND_UNITf)
				}

				obj.addLights(client.viewportLighting, client.viewportCellX, client.viewportCellY)
			}
		}
	}

	override fun thinkInner() {
		val entities = ObjectArrayList(entities)

		for (ent in entities) {
			ent.think()
		}

		val objects = ObjectArrayList(objects)

		for (ent in objects) {
			ent.thinkShared()
			ent.thinkClient()
		}
	}

	companion object {
		val ring = listOf(
			Vector2i(0, 0),
			Vector2i(1, 0),
			Vector2i(1, 1),
			Vector2i(1, -1),
			Vector2i(-1, 0),
			Vector2i(-1, 1),
			Vector2i(-1, -1),
		)
	}
}
