package ru.dbotthepony.kstarbound.defs.image

import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.io.json.consumeNull
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.SBPattern

class SpriteReference private constructor(
	val raw: AssetPath,
	val imagePath: SBPattern,
	val spritePath: SBPattern?,
	val image: Image?,
) {
	val sprite by lazy {
		if (image == null)
			null
		else if (spritePath == null)
			image.first
		else
			image[spritePath.value ?: return@lazy null]
	}

	fun with(values: (String) -> String?): SpriteReference {
		val imagePath = this.imagePath.with(values)
		val spritePath = this.spritePath?.with(values)

		if (imagePath != this.imagePath || spritePath != this.spritePath) {
			if (imagePath != this.imagePath) {
				val resolved = imagePath.value

				if (resolved == null)
					return SpriteReference(raw, imagePath, spritePath, null)
				else
					return SpriteReference(raw, imagePath, spritePath, Image.get(resolved))
			} else {
				return SpriteReference(raw, imagePath, spritePath, image)
			}
		}

		return this
	}

	fun with(values: Map<String, String>): SpriteReference {
		return with(values::get)
	}

	override fun equals(other: Any?): Boolean {
		return other is SpriteReference && other.imagePath == imagePath && other.spritePath == spritePath && other.image == image
	}

	override fun hashCode(): Int {
		return imagePath.hashCode() * 31 + spritePath.hashCode()
	}

	override fun toString(): String {
		return "SpriteReference[$imagePath:$spritePath]"
	}

	companion object : TypeAdapter<SpriteReference>() {
		val NEVER = SpriteReference(AssetPath("", ""), SBPattern.EMPTY, null, null)

		private val strings by lazy { Starbound.gson.getAdapter(String::class.java) }

		override fun write(out: JsonWriter, value: SpriteReference?) {
			out.value(value?.raw?.fullPath)
		}

		@JvmStatic
		fun create(path: String): SpriteReference {
			if (path == "")
				return NEVER

			val split = path.split(':')

			if (split.size > 2) {
				throw JsonSyntaxException("Ambiguous image reference: $path")
			}

			val imagePath = if (split.size == 2) SBPattern.of(split[0]) else SBPattern.of(path)
			val spritePath = if (split.size == 2) SBPattern.of(split[1]) else null

			if (imagePath.isPlainString) {
				val remapped = AssetPathStack.remap(split[0])
				return SpriteReference(AssetPath(path, AssetPathStack.remap(path)), SBPattern.raw(remapped), spritePath, Image.get(remapped))
			} else {
				return SpriteReference(AssetPath(path, path), imagePath, spritePath, null)
			}
		}

		override fun read(`in`: JsonReader): SpriteReference? {
			if (`in`.consumeNull())
				return null

			return create(strings.read(`in`))
		}
	}
}
