package ru.dbotthepony.kstarbound.world

import com.google.common.base.Predicate
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import ru.dbotthepony.kbox2d.api.ContactImpulse
import ru.dbotthepony.kbox2d.api.IContactFilter
import ru.dbotthepony.kbox2d.api.IContactListener
import ru.dbotthepony.kbox2d.api.Manifold
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kbox2d.dynamics.B2World
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.world.api.ICellAccess
import ru.dbotthepony.kstarbound.world.api.IChunkCell
import ru.dbotthepony.kstarbound.world.api.TileView
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kstarbound.world.entities.WorldObject
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.arrays.Object2DArray
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector2i
import java.lang.ref.ReferenceQueue
import java.lang.ref.WeakReference
import java.util.random.RandomGenerator

const val CHUNK_SIZE_BITS = 5
const val CHUNK_SIZE_MASK = 1 or 2 or 4 or 8 or 16
const val CHUNK_SIZE = 1 shl CHUNK_SIZE_BITS // 32
const val CHUNK_SIZE_FF = CHUNK_SIZE - 1
const val CHUNK_SIZEd = CHUNK_SIZE.toDouble()

@Suppress("UNCHECKED_CAST")
abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(
	val seed: Long,
	val size: Vector2i?,
	val loopX: Boolean,
	val loopY: Boolean
) : ICellAccess {
	val x: CoordinateMapper = if (size == null) CoordinateMapper.Infinite else if (loopX) CoordinateMapper.Wrapper(size.x) else CoordinateMapper.Clamper(size.x)
	val y: CoordinateMapper = if (size == null) CoordinateMapper.Infinite else if (loopY) CoordinateMapper.Wrapper(size.y) else CoordinateMapper.Clamper(size.y)

	// whenever provided cell position is within actual world borders, ignoring wrapping
	fun inBounds(x: Int, y: Int) = this.x.inBoundsCell(x) && this.y.inBoundsCell(y)
	fun inBounds(value: IStruct2i) = this.x.inBoundsCell(value.component1()) && this.y.inBoundsCell(value.component2())

	fun chunkFromCell(x: Int, y: Int) = ChunkPos(this.x.chunkFromCell(x), this.y.chunkFromCell(y))
	fun chunkFromCell(x: Double, y: Double) = ChunkPos(this.x.chunkFromCell(x.toInt()), this.y.chunkFromCell(y.toInt()))
	fun chunkFromCell(value: IStruct2i) = chunkFromCell(value.component1(), value.component2())
	fun chunkFromCell(value: IStruct2d) = chunkFromCell(value.component1(), value.component2())

	val background = TileView.Background(this)
	val foreground = TileView.Foreground(this)

	final override fun randomLongFor(x: Int, y: Int) = super.randomLongFor(x, y) xor seed

	override fun getCellDirect(x: Int, y: Int): IChunkCell? {
		if (!this.x.inBoundsCell(x) || !this.y.inBoundsCell(y)) return null
		return getCell(x, y)
	}

	override fun getCell(x: Int, y: Int): IChunkCell? {
		return chunkMap.getCell(x, y)
	}

	abstract inner class ChunkMap {
		abstract operator fun get(x: Int, y: Int): ChunkType?

		abstract fun promote(self: ChunkType)
		abstract fun purge()
		abstract fun remove(x: Int, y: Int)

		abstract fun getCell(x: Int, y: Int): IChunkCell?

		protected val queue = ReferenceQueue<ChunkType>()
		operator fun get(pos: ChunkPos) = get(pos.x, pos.y)

		protected inner class Ref(chunk: ChunkType) : WeakReference<ChunkType>(chunk, queue) {
			val pos = chunk.pos
		}

		protected fun create(x: Int, y: Int): ChunkType {
			purge()
			val pos = ChunkPos(x, y)
			val chunk = chunkFactory(pos)
			val orphanedInThisChunk = ArrayList<Entity>()

			for (ent in orphanedEntities) {
				val (ex, ey) = ent.position

				if (this@World.x.chunkFromCell(ex) == x && this@World.y.chunkFromCell(ey) == y) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			return chunk
		}
	}

	// hash chunk map is around 30% slower than rectangular one
	inner class HashChunkMap : ChunkMap() {
		private val map = Long2ObjectOpenHashMap<Any>()

		override fun getCell(x: Int, y: Int): IChunkCell? {
			if (!this@World.x.isValidCellIndex(x) || !this@World.y.isValidCellIndex(y)) return null
			val ix = this@World.x.cell(x)
			val iy = this@World.y.cell(y)
			return this[this@World.x.chunkFromCell(ix), this@World.y.chunkFromCell(iy)]?.getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK)
		}

		override fun get(x: Int, y: Int): ChunkType? {
			if (!this@World.x.isValidChunkIndex(x) || !this@World.y.isValidChunkIndex(y)) return null

			val x = this@World.x.chunk(x)
			val y = this@World.y.chunk(y)

			return map[ChunkPos.toLong(x, y)]?.let {
				if (it is World<*, *>.ChunkMap.Ref) {
					it.get() as ChunkType?
				} else {
					it as ChunkType?
				}
			} ?: create(x, y).also { map[ChunkPos.toLong(x, y)] = Ref(it) }
		}

		override fun purge() {
			var next = queue.poll() as World<*, *>.ChunkMap.Ref?

			while (next != null) {
				val k = ChunkPos.toLong(next.pos.x, next.pos.y)
				val get = map[k]

				if (get === next)
					map.remove(k)

				next = queue.poll() as World<*, *>.ChunkMap.Ref?
			}
		}

		override fun promote(self: ChunkType) {
			val (x, y) = self.pos
			val ref = map[ChunkPos.toLong(x, y)]

			if (ref !is World<*, *>.ChunkMap.Ref) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but there is $ref")
			}

			if (!(ref as World<*, ChunkType>.ChunkMap.Ref).refersTo(self)) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but it doesn't refer to valid object (self: $self, referent: ${ref.get()})")
			}

			map[ChunkPos.toLong(x, y)] = self
		}

		override fun remove(x: Int, y: Int) {
			val x = this@World.x.chunk(x)
			val y = this@World.y.chunk(y)

			val ref = map.remove(ChunkPos.toLong(x, y))

			if (ref is World<*, *>.ChunkMap.Ref) {
				ref.clear()
			}
		}
	}

	inner class ArrayChunkMap : ChunkMap() {
		val width = size!!.x
		val height = size!!.y

		private val map = Object2DArray.nulls<Any>(divideUp(width, CHUNK_SIZE), divideUp(height, CHUNK_SIZE))

		private fun getRaw(x: Int, y: Int): ChunkType {
			return map[x, y]?.let {
				if (it is World<*, *>.ChunkMap.Ref) {
					it.get() as ChunkType?
				} else {
					it as ChunkType?
				}
			} ?: create(x, y).also {
				map[x, y] = Ref(it)
			}
		}

		override fun getCell(x: Int, y: Int): IChunkCell? {
			if (!this@World.x.isValidCellIndex(x) || !this@World.y.isValidCellIndex(y)) return null

			val ix = this@World.x.cell(x)
			val iy = this@World.y.cell(y)

			return getRaw(ix ushr CHUNK_SIZE_BITS, iy ushr CHUNK_SIZE_BITS).getCell(ix and CHUNK_SIZE_MASK, iy and CHUNK_SIZE_MASK)
		}

		override fun get(x: Int, y: Int): ChunkType? {
			if (!this@World.x.isValidChunkIndex(x) || !this@World.y.isValidChunkIndex(y)) return null
			return getRaw(this@World.x.chunk(x), this@World.y.chunk(y))
		}

		override fun purge() {
			while (queue.poll() != null) {}
		}

		override fun promote(self: ChunkType) {
			val (x, y) = self.pos
			val ref = map[x, y]

			if (ref !is World<*, *>.ChunkMap.Ref) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but there is $ref")
			}

			if (!(ref as World<*, ChunkType>.ChunkMap.Ref).refersTo(self)) {
				throw IllegalStateException("Tried to promote chunk from weak to strong storage at $x, $y; but it doesn't refer to valid object (self: $self, referent: ${ref.get()})")
			}

			map[x, y] = self
		}

		override fun remove(x: Int, y: Int) {
			val x = this@World.x.chunk(x)
			val y = this@World.y.chunk(y)

			val old = map[x, y]

			if (old is World<*, *>.ChunkMap.Ref) {
				old.clear()
			}

			map[x, y] = null
		}
	}

	val chunkMap: ChunkMap = if (size != null) ArrayChunkMap() else HashChunkMap()

	val random: RandomGenerator = RandomGenerator.of("Xoroshiro128PlusPlus")

	/**
	 * Chunks, which have their collision mesh changed
	 */
	val dirtyPhysicsChunks = ObjectOpenHashSet<ChunkType>()

	val physics = B2World(Vector2d(0.0, -EARTH_FREEFALL_ACCELERATION))

	init {
		physics.contactFilter = object : IContactFilter {
			override fun shouldCollide(fixtureA: B2Fixture, fixtureB: B2Fixture): Boolean {
				val dataA = fixtureA.body!!.userData
				val dataB = fixtureB.body!!.userData

				//if (dataA is AbstractProjectileMovementController && dataB is AbstractProjectileMovementController) {
				//	return false
				//}

				return true
			}
		}

		physics.contactListener = object : IContactListener {
			override fun beginContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.beginContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.beginContact(contact)
				}
			}

			override fun endContact(contact: AbstractContact) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.endContact(contact)
				}

				if (dataB is IContactListener) {
					dataB.endContact(contact)
				}
			}

			override fun preSolve(contact: AbstractContact, oldManifold: Manifold) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.preSolve(contact, oldManifold)
				}

				if (dataB is IContactListener) {
					dataB.preSolve(contact, oldManifold)
				}
			}

			override fun postSolve(contact: AbstractContact, impulse: ContactImpulse) {
				val dataA = contact.fixtureA.body!!.userData
				val dataB = contact.fixtureB.body!!.userData

				if (dataA is IContactListener) {
					dataA.postSolve(contact, impulse)
				}

				if (dataB is IContactListener) {
					dataB.postSolve(contact, impulse)
				}
			}
		}
	}

	var ticks = 0
		private set

	fun think() {
		try {
			for (chunk in dirtyPhysicsChunks)
				chunk.bakeCollisions()

			dirtyPhysicsChunks.clear()
			physics.step(Starbound.TICK_TIME_ADVANCE, 6, 4)
			thinkInner()
			ticks++
		} catch(err: Throwable) {
			throw RuntimeException("Ticking world $this", err)
		}
	}

	protected abstract fun thinkInner()

	/**
	 * Сущности, у которых нет чанка, но они находятся в этом мире
	 */
	val orphanedEntities = ReferenceOpenHashSet<Entity>()

	/**
	 * Сущности, находящиеся в этом мире
	 */
	val entities = ReferenceLinkedOpenHashSet<Entity>()

	val objects = ReferenceLinkedOpenHashSet<WorldObject>()

	/**
	 * Стандартное ускорение свободного падения в Starbound Units/секунда^2
	 *
	 * При Vector2d.ZERO = невесомость
	 */
	var gravity by physics::gravity

	protected abstract fun chunkFactory(pos: ChunkPos): ChunkType

	fun testSpace(aabb: AABB, predicate: Predicate<IChunkCell> = Predicate { it.foreground.material != null }): Boolean {
		val tiles = aabb.encasingIntAABB()

		for (x in tiles.mins.x .. tiles.maxs.x) {
			for (y in tiles.mins.y .. tiles.maxs.y) {
				if (predicate.test(getCell(x, y) ?: continue)) {
					return true
				}
			}
		}

		return false
	}
}
