package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kbox2d.api.ContactEdge
import ru.dbotthepony.kbox2d.api.FixtureDef
import ru.dbotthepony.kbox2d.api.b2_polygonRadius
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.times
import kotlin.math.absoluteValue

enum class Move {
	STAND_STILL,
	MOVE_LEFT,
	MOVE_RIGHT
}

/**
 * Базовый абстрактный класс, реализующий сущность, которая ходит по земле
 */
abstract class WalkableMovementController<T : AliveWalkingEntity>(entity: T) : MovementController<T>(entity) {
	init {
		body.isFixedRotation = true
	}

	protected abstract val moveDirection: Move

	protected var sensorA: B2Fixture? = null
	protected var sensorB: B2Fixture? = null
	protected var bodyFixture: B2Fixture? = null

	protected abstract fun recreateBodyFixture()

	protected open fun recreateSensors() {
		sensorA?.destroy()
		sensorB?.destroy()

		val bodyFixture = bodyFixture ?: return
		val aabb = bodyFixture.shape.computeAABB(0)

		val sensorheight = (aabb.height - stepSize) / 2.0

		sensorA = body.createFixture(FixtureDef(
			shape = PolygonShape().also { it.setAsBox(0.2, sensorheight, Vector2d(-aabb.width / 2.0 - 0.2, aabb.height / 2.0 - sensorheight), 0.0) },
			isSensor = true,
		))

		sensorB = body.createFixture(FixtureDef(
			shape = PolygonShape().also { it.setAsBox(0.2, sensorheight, Vector2d(aabb.width / 2.0 + 0.2, aabb.height / 2.0 - sensorheight), 0.0) },
			isSensor = true,
		))
	}

	var wantsToDuck = false
	open var isDucked = false
		protected set

	override fun think() {
		super.think()
		thinkMovement()
	}

	/**
	 * See [IWalkableEntity.topSpeed]
	 */
	open val topSpeed by entity::topSpeed

	/**
	 * See [IWalkableEntity.moveSpeed]
	 */
	open val moveSpeed by entity::moveSpeed

	/**
	 * See [IWalkableEntity.freeFallMoveSpeed]
	 */
	open val freeFallMoveSpeed by entity::freeFallMoveSpeed

	/**
	 * See [IWalkableEntity.brakeForce]
	 */
	open val brakeForce by entity::brakeForce

	/**
	 * See [IWalkableEntity.stepSize]
	 */
	open val stepSize by entity::stepSize

	/**
	 * See [IWalkableEntity.jumpForce]
	 */
	open val jumpForce by entity::jumpForce

	protected var jumpRequested = false
	protected var nextJump = 0

	open fun requestJump() {
		if (jumpRequested || nextJump > world.ticks)
			return

		jumpRequested = true
		nextJump = world.ticks + 15
	}

	open fun recallJump() {
		if (jumpRequested) {
			jumpRequested = false
			nextJump = 0
		}
	}

	protected abstract fun canUnDuck(): Boolean

	protected var previousVelocity = Vector2d.ZERO

	protected open fun thinkMovement() {
		if (onGround && !isDucked) {
			when (moveDirection) {
				Move.STAND_STILL -> {
					body.linearVelocity += Vector2d(x = -body.linearVelocity.x * Starbound.TICK_TIME_ADVANCE * brakeForce)
				}

				Move.MOVE_LEFT -> {
					if (body.linearVelocity.x > 0.0) {
						body.linearVelocity += Vector2d(x = -body.linearVelocity.x * Starbound.TICK_TIME_ADVANCE * brakeForce)
					}

					if (body.linearVelocity.x > -topSpeed) {
						body.linearVelocity += Vector2d(x = -moveSpeed * Starbound.TICK_TIME_ADVANCE)

						// Ghost collision prevention
						if (body.linearVelocity.x.absoluteValue < 1) {
							body.linearVelocity += -Starbound.TICK_TIME_ADVANCE * world.gravity * 2.0
						}
					}

					var wantToStepUp = false
					var foundContact: ContactEdge? = null

					for (contact in body.contactEdgeIterator) {
						if (contact.contact.manifold.localNormal.dot(Vector2d.NEGATIVE_X) >= 0.95) {
							// we hit something to our left
							wantToStepUp = true
							foundContact = contact
							break
						}
					}

					if (wantToStepUp) {
						// make sure something we hit is actually a staircase of sort, and not geometry edges
						val aabbFound: AABB

						if (foundContact!!.contact.fixtureB == sensorA) {
							aabbFound = foundContact.contact.fixtureA.getAABB(foundContact.contact.childIndexA)
						} else {
							aabbFound = foundContact.contact.fixtureB.getAABB(foundContact.contact.childIndexB)
						}

						if (aabbFound.maxs.y - b2_polygonRadius * 2.0 <= body.worldSpaceAABB.mins.y) {
							// just a crack on sidewalk
							wantToStepUp = false
						}
					}

					if (wantToStepUp) {
						var stepHeightClear = true

						for (contact in body.contactEdgeIterator) {
							if (contact.contact.fixtureA == sensorA || contact.contact.fixtureB == sensorA) {
								if (contact.contact.isTouching) {
									stepHeightClear = false
									break
								}
							}
						}

						if (stepHeightClear) {
							val velocity = if (previousVelocity.length > body.linearVelocity.length) previousVelocity else body.linearVelocity
							body.setTransform(body.position + Vector2d(x = -0.05, y = stepSize), body.angle)
							body.linearVelocity = velocity
							//body.linearVelocity += Vector2d(y = -Starbound.SECONDS_IN_FRAME * 18.0 * stepSize * world.gravity.y)
						}
					}
				}

				Move.MOVE_RIGHT -> {
					if (body.linearVelocity.x < 0.0) {
						body.linearVelocity += Vector2d(x = -body.linearVelocity.x * Starbound.TICK_TIME_ADVANCE * brakeForce)
					}

					if (body.linearVelocity.x < topSpeed) {
						body.linearVelocity += Vector2d(x = moveSpeed * Starbound.TICK_TIME_ADVANCE)

						// Ghost collision prevention
						if (body.linearVelocity.x.absoluteValue < 1) {
							body.linearVelocity += -Starbound.TICK_TIME_ADVANCE * world.gravity * 2.0
						}
					}

					var wantToStepUp = false
					var foundContact: ContactEdge? = null

					for (contact in body.contactEdgeIterator) {
						if (contact.contact.manifold.localNormal.dot(Vector2d.POSITIVE_X) >= 0.95) {
							// we hit something to our right
							wantToStepUp = true
							foundContact = contact
							break
						}
					}

					if (wantToStepUp) {
						// make sure something we hit is actually a staircase of sort, and not geometry edges
						val aabbFound: AABB

						if (foundContact!!.contact.fixtureB == sensorB) {
							aabbFound = foundContact.contact.fixtureA.getAABB(foundContact.contact.childIndexA)
						} else {
							aabbFound = foundContact.contact.fixtureB.getAABB(foundContact.contact.childIndexB)
						}

						if (aabbFound.maxs.y - b2_polygonRadius * 2.0 <= body.worldSpaceAABB.mins.y) {
							// just a crack on sidewalk
							wantToStepUp = false
						}
					}

					if (wantToStepUp) {
						var stepHeightClear = true

						for (contact in body.contactEdgeIterator) {
							if (contact.contact.fixtureA == sensorB || contact.contact.fixtureB == sensorB) {
								if (contact.contact.isTouching) {
									stepHeightClear = false
									break
								}
							}
						}

						if (stepHeightClear) {
							val velocity = if (previousVelocity.length > body.linearVelocity.length) previousVelocity else body.linearVelocity
							body.setTransform(body.position + Vector2d(x = 0.05, y = stepSize), body.angle)
							body.linearVelocity = velocity
							//body.linearVelocity += Vector2d(y = -Starbound.SECONDS_IN_FRAME * 18.0 * stepSize * world.gravity.y)
						}
					}
				}
			}

			previousVelocity = body.linearVelocity

			if (jumpRequested) {
				jumpRequested = false
				nextJump = world.ticks + 15

				body.linearVelocity += Vector2d(y = jumpForce)
			}
		} else if (!onGround && !isDucked && freeFallMoveSpeed != 0.0) {
			when (moveDirection) {
				Move.STAND_STILL -> {
					// do nothing
				}

				Move.MOVE_LEFT -> {
					body.linearVelocity += Vector2d(x = -freeFallMoveSpeed * Starbound.TICK_TIME_ADVANCE)
				}

				Move.MOVE_RIGHT -> {
					body.linearVelocity += Vector2d(x = freeFallMoveSpeed * Starbound.TICK_TIME_ADVANCE)
				}
			}
		} else if (onGround && isDucked) {
			body.linearVelocity += Vector2d(x = -body.linearVelocity.x * Starbound.TICK_TIME_ADVANCE * brakeForce)
		}

		if (wantsToDuck && onGround) {
			isDucked = true
			recreateBodyFixture()
			recreateSensors()
			body.isAwake = true
		} else if (isDucked) {
			if (canUnDuck()) {
				isDucked = false
				recreateBodyFixture()
				recreateSensors()
				body.isAwake = true
			}
		}
	}
}

abstract class AliveEntity(world: World<*, *>) : Entity(world) {
	open var maxHealth = 10.0
	open var health = 10.0
}

abstract class AliveWalkingEntity(world: World<*, *>) : AliveEntity(world) {
	abstract override val movement: WalkableMovementController<*>

	/**
	 * Максимальная скорость передвижения этого AliveMovementController в Starbound Units/секунда
	 *
	 * Скорость передвижения: Это скорость вдоль земли (или в воздухе, если парит) при ходьбе.
	 *
	 * Если вектор скорости вдоль поверхности (или в воздухе, если парит) больше заданного значения,
	 * то сущность быстро тормозит (учитывая силу трения)
	 */
	open val topSpeed: Double = 20.0

	/**
	 * Скорость ускорения сущности в Starbound Units/секунда^2
	 *
	 * Если сущность хочет двигаться вправо или влево,
	 * то она разгоняется с данной скоростью.
	 */
	open val moveSpeed: Double = 64.0

	/**
	 * То, как сущность может влиять на свою скорость в Starbound Units/секунда^2
	 * когда находится в свободном падении или в состоянии невесомости.
	 *
	 * Позволяет в т.ч. игрокам изменять свою траекторию полёта в стиле Quake.
	 */
	open val freeFallMoveSpeed: Double = 8.0

	/**
	 * "Сила", с которой сущность останавливается, если не хочет двигаться.
	 *
	 * Зависит от текущего трения, так как технически является множителем трения поверхности,
	 * на которой стоит сущность.
	 */
	open val brakeForce: Double = 16.0

	/**
	 * Высота шага данной сущности. Данное значение отвечает за то, на сколько блоков
	 * сможет подняться сущность просто двигаясь в одном направлении без необходимости прыгнуть.
	 */
	open val jumpForce: Double = 20.0

	/**
	 * Импульс прыжка данной сущности. Если сущность парит, то данное значение не несёт никакой
	 * полезной нагрузки.
	 */
	open val stepSize: Double = 1.1

	open var wantsToDuck
		get() = movement.wantsToDuck
		set(value) { movement.wantsToDuck = value }

	open val isDucked get() = movement.isDucked
}
